import logging


class CeleryTaskFilter(logging.Filter):
    def filter(self, record):
        return record.processName.find("Worker") != -1


class CeleryProcessFilter(logging.Filter):
    def filter(self, record):
        return record.processName == "MainProcess"


class NotCeleryFilter(logging.Filter):
    def filter(self, record):
        return record.processName != "MainProcess" and record.processName.find("Worker") == -1



#!moo verb @password --on $player

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is a player command used to change the player's password.
"""

print("@password is not yet implemented.")



#!moo verb @gag at_gag --on $player --dspec any

# pylint: disable=return-outside-function,undefined-variable

"""
The concept of gagging is fairly simple to understand. If there is a player or object you do not wish to see any output
from, then you can place the player or object into your gag list. If any output is sent to you from that object, using
its `tell` verb, then it is ignored, and not printed on your terminal screen.

Two gaglists are maintained: one for players (in the property `gaglist`) and one for objects (in the property
`object_gaglist`).

Three verbs are used to maintain and observe the list of objects that are in the gag lists. The `@gag` verb is used as
a player command to add objects to the gag lists.
"""

from moo.core import context, lookup

parser = context.parser
player = context.player

# # this doesn't appear to be necessary
# if player != this:
#     player.tell("Permission denied.")
#     return

victims = [lookup(arg) for arg in args] if args else [parser.get_dobj(lookup=True)]

if not victims:
    print("Usage:  @gag <player or object> [<player or object>...]")
    return

gagplayers = []
gagobjs = []
for i, victim in enumerate(victims):
    if victim.is_player():
        gagplayers.append(victim)
    else:
        gagobjs.append(victim)

changed = False
for p in gagplayers:
    if p in player.gaglist:
        print(f"You are already gagging {p.name}.")
    else:
        changed = True
        tmp = player.gaglist
        tmp.append(p)
        player.gaglist = tmp

for o in gagobjs:
    if o in player.object_gaglist:
        print(f"You are already gagging {o.name}.")
    else:
        changed = True
        tmp = player.object_gaglist
        tmp.append(o)
        player.object_gaglist = tmp

if changed:
    print("Gag list updated.")
else:
    print("No changes made to gag list.")



#!moo verb set_name --on $root_class --dspec any

# pylint: disable=return-outside-function,undefined-variable

"""
This verb sets the name of the object to value. It returns `True` if the name was set to the value successfully,
otherwise it returns `False`. This verb and the `title` verb are used to control access to the name property of an
object.
"""

try:
    this.name = args[0]
    this.save()
except:  # pylint: disable=bare-except
    return False

return True



#!moo verb recycle --on $room

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is called by the LambdaMOO server when a room is recycled. When a room is recycled, something has to be done
with the room's contents, both players and objects, to stop them ending up in `$nothing`. This is done by trying to move
everything home.

The main loop goes through the complete contents list of the room. If the object is a player, then it is moved home,
using it's `moveto` verb, or to the `$player_start` room, using the `move()` primitive. All other objects are moved to the
inventories of their owners.

Note that if the attempt to move an object fails, then no action is taken. The server will place all objects still in
the room into `$nothing` when the room is recycled.
"""

for x in this.contents.all():
    if x.is_player():
        if x.has_property("home"):
            x.moveto(x.home)
        if x.location == this:
            x.location = _.player_start
            x.save()
    else:
        x.moveto(x.owner)



# -*- coding: utf-8 -*-
"""
The primary Object class
"""

import json
import logging
from typing import Generator

from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models.query import Q, QuerySet
from django.db.models.signals import m2m_changed
from django.dispatch import receiver

from moo import bootstrap
from .. import exceptions, invoke, utils
from ..code import ContextManager
from .acl import Access, AccessibleMixin, Permission
from .auth import Player
from .property import Property
from .verb import Verb, Preposition, PrepositionName, PrepositionSpecifier, VerbName

log = logging.getLogger(__name__)


@receiver(m2m_changed)
def relationship_changed(sender, instance, action, model, signal, reverse, pk_set, using, **kwargs):
    child = instance
    if not (sender is Relationship and not reverse):
        return
    elif action in ("pre_add", "pre_remove"):
        child.can_caller("transmute", instance)
        for pk in pk_set:
            parent = model.objects.get(pk=pk)
            parent.can_caller("derive", parent)
        return
    elif action != "post_add":
        return
    for pk in pk_set:
        parent = model.objects.get(pk=pk)
        # pylint: disable=redefined-builtin
        for property in Property.objects.filter(origin=parent):
            if property.inherit_owner:
                new_owner = property.owner
            else:
                new_owner = child.owner
            Property.objects.update_or_create(
                name=property.name,
                origin=child,
                defaults=dict(
                    owner=new_owner,
                    inherit_owner=property.inherit_owner,
                ),
                create_defaults=dict(
                    owner=new_owner,
                    inherit_owner=property.inherit_owner,
                    value=property.value,
                    type=property.type,
                ),
            )


class Object(models.Model, AccessibleMixin):
    #: The canonical name of the object
    name = models.CharField(max_length=255, db_index=True)
    #: If True, this object is the only object with this name
    unique_name = models.BooleanField(default=False, db_index=True)
    #: This object should be obvious among a group. The meaning of this value is database-dependent.
    obvious = models.BooleanField(default=True)
    #: The owner of this object. Changes require `entrust` permission.
    owner = models.ForeignKey("self", related_name="+", blank=True, null=True, on_delete=models.SET_NULL)
    parents = models.ManyToManyField(
        "self",
        related_name="children",
        blank=True,
        symmetrical=False,
        through="Relationship",
    )
    """
    The parents of this object. Changes require `derive` and `transmute` permissions, respectively.

    .. code-block:: Python

        from moo.core import context, lookup
        # in the default DB, all wizards inherit from this Object
        wizard_class = lookup("wizard class")
        # Changes to ManyToMany fields like this are automatically saved
        context.caller.parents.add(wizard_class)
    """
    location = models.ForeignKey(
        "self",
        related_name="contents",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        db_index=True,
    )
    """
    The location of this object. When changing, this kicks off some other verbs:

    If `where` is the new object, then the verb-call `where.accept(self)` is performed before any movement takes place.
    If the verb returns a false value and the programmer is not a wizard, then `where` is considered to have refused entrance
    to `self` and raises :class:`.PermissionError`. If `where` does not define an `accept` verb, then it is treated as if
    it defined one that always returned false.

    If moving `what` into `self` would create a loop in the containment hierarchy (i.e., what would contain itself, even
    indirectly), then :class:`.RecursiveError` is raised instead.

    Let `old` be the location of `self` before it was moved. If `old` is a valid object, then the verb-call
    `old.exitfunc(self)` is performed and its result is ignored; it is not an error if `old` does not define
    a verb named `exitfunc`.

    Finally, if `where` is still the location of `self`, then the verb-call `where.enterfunc(self)` is performed and its
    result is ignored; again, it is not an error if `where` does not define a verb named `enterfunc`.
    """

    @classmethod
    def from_db(cls, db, field_names, values):
        instance = super().from_db(db, field_names, values)
        instance.original_owner = values[field_names.index("owner_id")]
        instance.original_location = values[field_names.index("location_id")]
        return instance

    def __str__(self):
        return "#%s (%s)" % (self.id, self.name)

    @property
    def kind(self):
        return "object"

    def is_player(self) -> bool:
        """
        Check if this object is a player avatar.
        """
        return Player.objects.filter(avatar=self).exists()

    def is_wizard(self) -> bool:
        """
        Check if this object is a player avatar.
        """
        return Player.objects.filter(avatar=self, wizard=True).exists()

    def is_connected(self) -> bool:
        """
        Check if this object is a player avatar that is currently connected.
        """
        return True

    def is_named(self, name: str) -> bool:
        """
        Check if this object has a name or alias that matches the given name.
        """
        for alias in self.aliases.all():
            if alias.alias.lower() == name.lower():
                return True
        return self.name.lower() == name.lower()  # pylint: disable=no-member

    def find(self, name: str) -> 'QuerySet["Object"]':
        """
        Find contents by the given name or alias.

        :param name: the name or alias to search for, case-insensitive
        """
        self.can_caller("read", self)
        qs = Object.objects.filter(location=self, name__iexact=name)
        aliases = Object.objects.filter(location=self, aliases__alias__iexact=name)
        return qs.union(aliases)

    def contains(self, obj: "Object"):
        for item in self.get_contents():
            if item == obj:
                return True
        return False

    def is_a(self, obj: "Object") -> bool:
        """
        Check if this object is a child of the provided object.

        :param obj: the potential parent object
        :return: True if this object is a child of `obj`
        """
        return obj in self.get_ancestors()

    def get_ancestors(self) -> Generator["Object", None, None]:
        """
        Get the ancestor tree for this object.
        """
        self.can_caller("read", self)
        # TODO: One day when Django 5.0 works with `django-cte` this can be SQL.
        for parent in self.parents.all():  # pylint: disable=no-member
            yield from parent.get_ancestors()
            yield parent

    def get_descendents(self) -> Generator["Object", None, None]:
        """
        Get the descendent tree for this object.
        """
        self.can_caller("read", self)
        # TODO: One day when Django 5.0 works with `django-cte` this can be SQL.
        for child in self.children.all():
            yield child
            yield from child.get_descendents()

    def get_contents(self) -> Generator["Object", None, None]:
        """
        Get the content tree for this object.
        """
        self.can_caller("read", self)
        # TODO: One day when Django 5.0 works with `django-cte` this can be SQL.
        for item in self.contents.all():
            yield item
            yield from item.get_contents()

    def add_verb(
        self,
        *names: list[str],
        code: str = None,
        owner: "Object" = None,
        repo=None,
        filename: str = None,
        direct_object: str = "none",
        indirect_objects: dict[str, str] = None,
        replace: bool = False,
    ):
        """
        Defines a new :class:`.Verb` on the given object.

        :param names: a list of names for the new verb
        :param code: the Python code for the new Verb
        :param owner: the owner of the Verb being created
        :param repo: optional, the Git repo this code is from
        :param filename: optional, the name of the code file within the repo
        :param direct_object: a direct object specifier for the verb
        :param indirect_objects: a list of indirect object specifiers for the verb
        """
        self.can_caller("write", self)
        owner = ContextManager.get("caller") or owner or self
        if filename and not code:
            code = bootstrap.get_source(filename, dataset=repo.slug)
        verb = None
        if replace and names:
            verb = Verb.objects.filter(origin=self, names__name=names[0]).first()
        if verb is not None:
            verb.owner = owner
            verb.repo = repo
            verb.filename = filename
            verb.code = code
            verb.direct_object = direct_object
            verb.save()
            verb.indirect_objects.clear()
            verb.names.all().delete()
        else:
            verb = Verb.objects.create(
                origin=self,
                owner=owner,
                repo=repo,
                filename=filename,
                code=code,
                direct_object=direct_object,
            )
        if indirect_objects is not None:
            for prep, specifier in indirect_objects.items():
                if prep in ["any", "none"]:
                    p = None
                    prep_specifier = prep
                else:
                    pn = PrepositionName.objects.get(name=prep)
                    p = pn.preposition
                    prep_specifier = "none"
                verb.indirect_objects.add(
                    PrepositionSpecifier.objects.update_or_create(
                        preposition=p, preposition_specifier=prep_specifier, specifier=specifier
                    )[0]
                )
        for name in names:
            for item in utils.expand_wildcard(name):
                verb.names.add(VerbName.objects.create(verb=verb, name=item))
        return verb

    def invoke_verb(self, name, *args, **kwargs):
        """
        Invoke a :class:`.Verb` defined on the given object, traversing the inheritance tree until it's found.

        :param name: the name of the verb
        :param args: positional arguments for the verb
        :param kwargs: keyword arguments for the verb
        """
        qs = self._lookup_verb(name, recurse=True)
        verb = qs[0]
        self.can_caller("execute", verb)
        verb.invoked_name = name
        verb.invoked_object = self
        return verb(*args, **kwargs)

    def has_verb(self, name, recurse=True):
        """
        Check if a particular :class:`.Verb` is defined on this object.

        :param name: the name of the verb
        :param recurse: whether or not to traverse the inheritance tree
        """
        self.can_caller("read", self)
        try:
            self._lookup_verb(name, recurse)
        except Verb.DoesNotExist:
            return False
        return True

    def get_verb(self, name, recurse=True, allow_ambiguous=False, return_first=True):
        """
        Retrieve a specific :class:`.Verb` instance defined on this Object.

        :param name: the name of the verb
        :param recurse: whether or not to traverse the inheritance tree
        :param return_first: if True, return the first matching verb, otherwise return all matching verbs
        """
        self.can_caller("read", self)
        verbs = self._lookup_verb(name, recurse, return_first)
        if len(verbs) > 1 and not allow_ambiguous:
            raise exceptions.AmbiguousVerbError(name, verbs)
        for v in verbs:
            v.invoked_name = name
            v.invoked_object = self
        if allow_ambiguous:
            return verbs
        v = verbs[0]
        return v

    def parse_verb(self, parser):
        """
        Check if this parser instance could refer to a verb on this object.
        """
        result = []
        for verb in self._lookup_verb(parser.words[0], recurse=True, return_first=False):
            if verb.direct_object == "this" and parser.dobj != self:
                continue
            if verb.direct_object == "none" and parser.has_dobj_str():
                continue
            if verb.direct_object == "any" and not parser.has_dobj_str():
                continue
            for ispec in verb.indirect_objects.all():
                for prep, values in parser.prepositions.items():
                    if ispec.preposition_specifier == "none":
                        continue
                    if ispec.preposition_specifier == "this" and values[2] != self:
                        continue
                    if ispec.preposition_specifier != "any":
                        if not ispec.preposition.names.filter(name=prep).exists():
                            continue
            # sometimes an object has multiple verbs with the same name after inheritance
            # so we need to check if the verb is already in the list
            if verb not in result:
                result.append(verb)
        if not result:
            return None
        if len(result) == 1:
            return result[0]
        raise exceptions.AmbiguousVerbError(parser.words[0], result)

    def _lookup_verb(self, name, recurse=True, return_first=True):
        found = []
        qs = Verb.objects.filter(origin=self, names__name=name)
        if not qs and recurse:
            for ancestor in reversed(list(self.get_ancestors())):
                qs = Verb.objects.filter(origin=ancestor, names__name=name)
                if qs:
                    if return_first:
                        return qs
                    else:
                        found.extend(qs.all())
        elif qs:
            if return_first:
                return qs
            else:
                found.extend(qs.all())
        if found:
            return found
        else:
            raise Verb.DoesNotExist(f"No such verb `{name}`.")

    def set_property(self, name, value, inherit_owner=False, owner=None):
        """
        Defines a new :class:`.Property` on the given object.

        :param names: a list of names for the new Property
        :param value: the value for the new Property
        :param inherit_owner: if True, this property's owner will be reassigned on child instances
        :param owner: the owner of the Property being created
        """
        from .. import moojson

        self.can_caller("write", self)
        owner = ContextManager.get("caller") or owner or self
        Property.objects.update_or_create(
            name=name,
            origin=self,
            defaults=dict(
                value=moojson.dumps(value),
                owner=owner,
                type="string",
                inherit_owner=inherit_owner,
            ),
        )

    def get_property(self, name, recurse=True, original=False):
        """
        Retrieve a :class:`.Property` instance defined on this Object.

        :param name: the name of the verb
        :param recurse: whether or not to traverse the inheritance tree
        :param original: if True, return the whole Property object, not just its value
        """
        from .. import moojson

        self.can_caller("read", self)
        qs = Property.objects.filter(origin=self, name=name)
        if not qs and recurse:
            for ancestor in reversed(list(self.get_ancestors())):
                qs = Property.objects.filter(origin=ancestor, name=name)
                if qs:
                    break
        if qs:
            return qs[0] if original else moojson.loads(qs[0].value)
        else:
            raise Property.DoesNotExist(f"No such property `{name}`.")

    def has_property(self, name, recurse=True):
        """
        Check if a particular :class:`.Property` is defined on this object.
        """
        self.can_caller("read", self)
        qs = Property.objects.filter(origin=self, name=name)
        if not qs and recurse:
            for ancestor in reversed(list(self.get_ancestors())):
                qs = Property.objects.filter(origin=ancestor, name=name)
                if qs:
                    break
        return qs.exists()

    def delete(self, *args, **kwargs):
        if self.has_verb("recycle", recurse=False):
            self.invoke_verb("recycle")
        try:
            quota = self.owner.get_property("ownership_quota", recurse=False)
            if quota is not None:
                self.owner.set_property("ownership_quota", quota + 1)
        except Property.DoesNotExist:
            pass
        super().delete(*args, **kwargs)

    def save(self, *args, **kwargs):
        unsaved = self.pk is None
        if unsaved:
            # there's no permissions yet, so we can't check for `entrust`
            caller = ContextManager.get("caller")
            if caller and self.owner != caller:
                raise PermissionError("Can't change owner at creation time.")
        super().save(*args, **kwargs)
        # after saving, new objects need default permissions
        if unsaved:
            utils.apply_default_permissions(self)
        # Recursion Check: note that this leaves a broken object behind unless run in a transaction
        if self.location and self.contains(self.location):
            raise exceptions.RecursiveError(f"{self} already contains {self.location}")
        # ACL Check: to change owner, caller must be allowed to `entrust` on this object
        original_owner = getattr(self, "original_owner", None)
        original_owner = Object.objects.get(pk=original_owner) if original_owner else None
        if original_owner != self.owner and self.owner:
            self.can_caller("entrust", self)
        # ACL Check: to change anything else about the object you at least need `write`
        self.can_caller("write", self)
        # ACL Check: to change the location, caller must be allowed to `move` on this object
        original_location = getattr(self, "original_location", None)
        original_location = Object.objects.get(pk=original_location) if original_location else None
        if original_location != self.location and self.location:
            self.can_caller("move", self)
            # the new location must define an `accept` verb that returns True for this obejct
            if self.location.has_verb("accept"):
                if not self.location.invoke_verb("accept", self):
                    raise PermissionError(f"{self.location} did not accept {self}")
            else:
                raise PermissionError(f"{self.location} did not accept {self}")
            # the optional `exitfunc` Verb will be called asyncronously
            if original_location and original_location.has_verb("exitfunc"):
                invoke(self, verb=original_location.get_verb("exitfunc"))
            # the optional `enterfun` Verb will be called asyncronously
            if self.location and self.location.has_verb("enterfunc"):
                invoke(self, verb=self.location.get_verb("enterfunc"))

    # Django gets upset if this meddles with anything in RESERVED_NAMES
    # but otherwise this seems to work, including in the admin interface
    def __getattr__(self, name):
        if name in RESERVED_NAMES:
            raise AttributeError(name)
        if self.has_verb(name, recurse=True):
            return self.get_verb(name, recurse=True)
        if self.has_property(name, recurse=True):
            return self.get_property(name, recurse=True)
        raise AttributeError(f"{self} has no attribute `{name}`")

    def owns(self, subject) -> bool:
        """
        Convenience method to check if the `subject` is owned by `self`
        """
        return subject.owner == self

    def is_allowed(self, permission: str, subject, fatal: bool = False) -> bool:
        """
        Check if this object is allowed to perform an action on an object.

        :param permission: the name of the permission to check
        :param subject: the item to check against
        :type subject: Union[Object, Verb, Property]
        :param fatal: if True, raise a :class:`.PermissionError` instead of returning False
        :raises PermissionError: if permission is denied and `fatal` is set to True
        """
        # Get both permissions in a single query
        perms = Permission.objects.filter(name__in=(permission, "anything")).values_list("id", flat=True)
        if len(perms) < 2:
            perms = list(perms) + [Permission.objects.get(name=permission).id]

        # Build subject filter once
        subject_filter = Q()
        if subject.kind == "object":
            subject_filter = Q(object=subject)
        elif subject.kind == "verb":
            subject_filter = Q(verb=subject)
        elif subject.kind == "property":
            subject_filter = Q(property=subject)

        # Build OR conditions for all matching rules
        query = subject_filter & Q(permission_id__in=perms) & (
            Q(type="accessor", accessor=self) |
            Q(type="group", group="everyone") |
            (Q(type="group", group="owners") if self.owns(subject) else Q()) |
            (Q(type="group", group="wizards") if Player.objects.filter(avatar=self, wizard=True).exists() else Q())
        )

        rules = Access.objects.filter(query).order_by("rule", "type")

        if rules:
            for rule in rules:
                if rule.rule == "deny":
                    if fatal:
                        raise PermissionError(f"{self} is explicitly denied {permission} on {subject}")
                    return False
            return True
        elif fatal:
            raise PermissionError(f"{self} is not allowed {permission} on {subject}")
        else:
            return False


# these are the name that django relies on __getattr__ for, there may be others
RESERVED_NAMES = [
    "resolve_expression",
    "get_source_expressions",
    "_prefetched_objects_cache",
    "original_owner",
    "original_location",
]


class Relationship(models.Model):
    class Meta:
        unique_together = [["child", "parent"]]

    child = models.ForeignKey(Object, related_name="+", on_delete=models.CASCADE)
    parent = models.ForeignKey(Object, related_name="+", on_delete=models.CASCADE)
    weight = models.IntegerField(default=0)


class Alias(models.Model):
    class Meta:
        verbose_name_plural = "aliases"

    object = models.ForeignKey(Object, related_name="aliases", on_delete=models.CASCADE)
    alias = models.CharField(max_length=255)

    def save(self, *args, **kwargs):
        self.object.can_caller("write", self.object)
        super().save(*args, **kwargs)



#!moo verb @exits --on $room

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is a player command used to print a list of the exits in a room. It can only be used by the owner of the
room. The verb simply runs through the list of defined exits, stored in the property exits, and prints the exit name,
object reference number, destination name, and exit aliases.
"""

exits = this.get_property("exits") or []
if not exits:
    print("[color red]There are no exits defined for this room.[/color red]")
    return

print("[color cyan]Exits defined for this room:[/color cyan]")
for exit in exits:
    exit_name = exit.name
    dest_name = exit.dest.name
    aliases = ", ".join([x.alias for x in exit.aliases.all()])
    print(f"- [color yellow]{exit_name}[/color yellow] (Aliases: {aliases}) "
          f"to [color green]{dest_name}[/color green] (#{exit.dest.id})")



#!moo verb look --on "player class" --dspec either

# pylint: disable=return-outside-function,undefined-variable

from moo.core import context

qs = context.caller.location.properties.filter(name="description")
if qs:
    print(qs[0].value)
else:
    print("No description.")



#!moo verb @sw*eep --on $player

# pylint: disable=return-outside-function,undefined-variable

"""
This is a player command used to list the objects that are capable of receiving any messages sent out from the player
in the current room. It gathers a list of the objects in the current room, from the `contents` property of the player's
location. If any element in the list is a connected player, the message

    blip (#42) is listening.

is printed, for example.

If an element has a `sweep_msg` verb defined, the returned string from this verb is printed, prepended by the object's
name.

If an element has a `tell` verb defined, and the owner is not the invoking player or a wizard, then this object is a
potential snooper, and is reported by a phrase such as:

    The Fruitbat (#999) has been taught to listen by blip (#42).

The verbs `announce', `announce_all', `announce_all_but', `say', `emote', `huh', `huh2' and `whisper' are checked to
see if the current room has a definition not owned by the invoking player or a wizard. If any of these are found, a
message such as:

    The Venue Hallway (#1234) may have been bugged by blip.

if the player's location was The Venue Hallway.

If no potential bugs are found, the the message `Communications are secure.' is printed, and the player can breathe
easily (ish).
"""
from moo.core import context

buggable_verbs = ["announce", "announce_all", "announce_all_but", "say", "emote", "huh", "huh2", "whisper"]

player = context.player
room = player.location
secure = True
for obj in room.contents.all():
    if obj.is_player():
        print(f"{obj} is listening.")
    if obj.has_verb("sweep_msg"):
        print(f"{obj} {obj.invoke_verb('sweep_msg')}")
    if obj.has_verb("tell"):
        tell = obj.get_verb("tell")
        if tell.owner != player and not tell.owner.is_wizard():
            secure = False
            print(f"{obj} has been taught to listen by {tell.owner}.")
already_printed = {}
for verb in buggable_verbs:
    if room.has_verb(verb):
        v = room.get_verb(verb)
        if v.owner != player and not v.owner.is_wizard():
            msg = f"{room} may have been bugged by {v.owner}."
            if msg not in already_printed:
                already_printed[msg] = True
                secure = False
                print(msg)
if secure:
    print("Communications are secure.")



import pytest

from moo.core import code, lookup, parse
from moo.core.models import Object


# --- message verbs ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_page_absent_msg(t_init: Object, t_wizard: Object):
    """page_absent_msg() returns a string containing the player's name."""
    with code.ContextManager(t_wizard, lambda _: None):
        result = t_wizard.page_absent_msg()
    assert "Wizard" in result


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_page_origin_msg(t_init: Object, t_wizard: Object):
    """page_origin_msg() returns a string with the player's name and location."""
    with code.ContextManager(t_wizard, lambda _: None):
        result = t_wizard.page_origin_msg()
    assert "Wizard" in result
    assert "The Laboratory" in result


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_page_echo_msg(t_init: Object, t_wizard: Object):
    """page_echo_msg() returns the echo confirmation string."""
    with code.ContextManager(t_wizard, lambda _: None):
        result = t_wizard.page_echo_msg()
    assert result == "Your message has been sent."


# --- page ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_page_sends_origin_msg(t_init: Object, t_wizard: Object):
    """page <player> sends the page_origin_msg to the recipient via tell."""
    with code.ContextManager(t_wizard, lambda _: None) as ctx:
        with pytest.warns(RuntimeWarning) as w:
            parse.interpret(ctx, "page Player")
    messages = [str(x.message) for x in w.list]
    assert any("Wizard" in m and "The Laboratory" in m for m in messages)


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_page_with_message(t_init: Object, t_wizard: Object):
    """page <player> with <message> sends the origin msg and the message to the recipient."""
    with code.ContextManager(t_wizard, lambda _: None) as ctx:
        with pytest.warns(RuntimeWarning) as w:
            parse.interpret(ctx, "page Player with Hello there")
    messages = [str(x.message) for x in w.list]
    assert any("Wizard" in m for m in messages)
    assert any('pages, "Hello there"' in m for m in messages)


# --- whisper ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_whisper_sends_message(t_init: Object, t_wizard: Object):
    """whisper <message> to <player> delivers the whisper to the recipient via write()."""
    with code.ContextManager(t_wizard, lambda _: None) as ctx:
        with pytest.warns(RuntimeWarning, match="whispers to you"):
            parse.interpret(ctx, "whisper Hello to Player")


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_whisper_no_caller_echo(t_init: Object, t_wizard: Object):
    """whisper delivers to the recipient but sends no echo to the caller."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        with pytest.warns(RuntimeWarning, match="whispers to you"):
            parse.interpret(ctx, "whisper Hello to Player")
    assert printed == []


# --- page ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_page_sends_caller_echo(t_init: Object, t_wizard: Object):
    """page sends the page_echo_msg confirmation back to the caller via print()."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        with pytest.warns(RuntimeWarning):
            parse.interpret(ctx, "page Player")
    assert "Your message has been sent." in printed


# --- announce ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_announce_delegates_to_room(t_init: Object, t_wizard: Object):
    """announce() delegates to the location's announce verb when it exists."""
    player_npc = lookup("Player")
    player_npc.location = t_wizard.location
    player_npc.save()
    with code.ContextManager(t_wizard, lambda _: None):
        with pytest.warns(RuntimeWarning) as w:
            t_wizard.announce("hello from wizard")
    messages = [str(x.message) for x in w.list]
    assert any("hello from wizard" in m for m in messages)



# Generated by Django 4.2.5 on 2023-11-14 01:57

import django.db.models.deletion
from django.db import migrations, models

import moo.core.models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Object",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255)),
                ("unique_name", models.BooleanField()),
                (
                    "location",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="contents",
                        to="core.object",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Permission",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="Repository",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("slug", models.SlugField()),
                ("url", moo.core.models.URLField(max_length=255)),
                ("prefix", models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="Verb",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("code", models.TextField(blank=True, null=True)),
                ("filename", models.CharField(blank=True, max_length=255, null=True)),
                ("ref", models.CharField(blank=True, max_length=255, null=True)),
                ("ability", models.BooleanField()),
                ("method", models.BooleanField()),
                (
                    "origin",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="verbs", to="core.object"
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="core.object",
                    ),
                ),
                (
                    "repo",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="core.repository",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="VerbName",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255)),
                (
                    "verb",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="names", to="core.verb"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Task",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("verb_name", models.CharField(max_length=255)),
                ("args", models.TextField()),
                ("kwargs", models.TextField()),
                ("created", models.DateTimeField()),
                ("delay", models.IntegerField()),
                ("killed", models.BooleanField(default=False)),
                ("error", models.CharField(blank=True, max_length=255, null=True)),
                ("trace", models.TextField(blank=True, null=True)),
                (
                    "origin",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name="+", to="core.object"),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="tasks", to="core.object"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Relationship",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("weight", models.IntegerField(default=0)),
                (
                    "child",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="parent", to="core.object"
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="child", to="core.object"
                    ),
                ),
            ],
            options={
                "unique_together": {("child", "parent")},
            },
        ),
        migrations.CreateModel(
            name="Property",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255)),
                ("value", models.TextField(blank=True, null=True)),
                (
                    "type",
                    models.CharField(
                        choices=[("string", "string"), ("python", "python"), ("dynamic", "dynamic")], max_length=255
                    ),
                ),
                (
                    "origin",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="properties", to="core.object"
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        null=True, on_delete=django.db.models.deletion.SET_NULL, related_name="+", to="core.object"
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "properties",
            },
        ),
        migrations.CreateModel(
            name="Player",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("session_id", models.CharField(blank=True, max_length=255, null=True)),
                ("wizard", models.BooleanField(default=False)),
                ("enabled", models.BooleanField(default=True)),
                ("crypt", models.CharField(blank=True, max_length=255, null=True)),
                ("last_login", models.DateTimeField(null=True)),
                ("last_logout", models.DateTimeField(null=True)),
                (
                    "avatar",
                    models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to="core.object"),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Observation",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "object",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="observer", to="core.object"
                    ),
                ),
                (
                    "observer",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="object", to="core.object"
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="object",
            name="observers",
            field=models.ManyToManyField(
                blank=True, related_name="observing", through="core.Observation", to="core.object"
            ),
        ),
        migrations.AddField(
            model_name="object",
            name="owner",
            field=models.ForeignKey(
                blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name="+", to="core.object"
            ),
        ),
        migrations.AddField(
            model_name="object",
            name="parents",
            field=models.ManyToManyField(
                blank=True, related_name="children", through="core.Relationship", to="core.object"
            ),
        ),
        migrations.CreateModel(
            name="Alias",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("alias", models.CharField(max_length=255)),
                (
                    "object",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="aliases", to="core.object"
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "aliases",
            },
        ),
        migrations.CreateModel(
            name="Access",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("rule", models.CharField(choices=[("allow", "allow"), ("deny", "deny")], max_length=5)),
                ("type", models.CharField(choices=[("accessor", "accessor"), ("group", "group")], max_length=8)),
                (
                    "group",
                    models.CharField(
                        choices=[("everyone", "everyone"), ("owners", "owners"), ("wizards", "wizards")],
                        max_length=8,
                        null=True,
                    ),
                ),
                ("weight", models.IntegerField(default=0)),
                (
                    "accessor",
                    models.ForeignKey(
                        null=True, on_delete=django.db.models.deletion.CASCADE, related_name="rights", to="core.object"
                    ),
                ),
                (
                    "object",
                    models.ForeignKey(
                        null=True, on_delete=django.db.models.deletion.CASCADE, related_name="acl", to="core.object"
                    ),
                ),
                (
                    "permission",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="usage", to="core.permission"
                    ),
                ),
                (
                    "property",
                    models.ForeignKey(
                        null=True, on_delete=django.db.models.deletion.CASCADE, related_name="acl", to="core.property"
                    ),
                ),
                (
                    "verb",
                    models.ForeignKey(
                        null=True, on_delete=django.db.models.deletion.CASCADE, related_name="acl", to="core.verb"
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "access controls",
            },
        ),
    ]



#!moo verb @entrances --on $room

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is a player command used to list the entrances to the player's current location. Only the owner of a room may
list it's entrances. For every object kept in the room's entrances list, the exit name, object reference number and
aliases are displayed to the player.
"""

entrances = this.get_property("entrances") or []
if not entrances:
    print("[color red]There are no entrances defined for this room.[/color red]")
    return

print("[color cyan]Entrances defined for this room:[/color cyan]")
for exit in entrances:
    exit_name = exit.name
    dest_name = exit.dest.name
    aliases = ", ".join([x.alias for x in exit.aliases.all()])
    print(f"- [color yellow]{exit_name}[/color yellow] (Aliases: {aliases}) "
          f"to [color green]{dest_name}[/color green] (#{exit.dest.id})")



#!moo verb accept --on $container

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is used to determine if a player will allow `thing` to be moved into his/her inventory. The verb defined for
the `$player` class allows anything that is not a player to be moved into the player's possession. You could override
this verb to restrict the sorts of things you would want other people to be able to place on your person.
"""

thing = args[0]
return not thing.is_player()



#!moo verb otake_succeeded_msg otake_failed_msg take_succeeded_msg take_failed_msg odrop_succeeded_msg odrop_failed_msg drop_succeeded_msg drop_failed_msg --on $thing

# pylint: disable=return-outside-function,undefined-variable

"""
These verbs return a pronoun substituted version of the corresponding properties stored on the thing object. They are
used by `$thing.take` and `$thing.drop`.
"""

from moo.core import context

prop_name = verb_name
prop_value = this.get_property(prop_name)
actor = context.player

if prop_name == 'take_succeeded_msg':
    return prop_value.format(actor="You", subject=this.title())
elif prop_name == 'otake_failed_msg':
    return prop_value.format(actor=actor, subject=this.title())
elif prop_name == 'take_succeeded_msg':
    return prop_value.format(actor="You", subject=this.title())
elif prop_name == 'take_failed_msg':
    return prop_value.format(actor="You", subject=this.title())
elif prop_name == 'odrop_succeeded_msg':
    return prop_value.format(actor=actor, subject=this.title())
elif prop_name == 'odrop_failed_msg':
    return prop_value.format(actor=actor, subject=this.title())
elif prop_name == 'drop_succeeded_msg':
    return prop_value.format(actor="You", subject=this.title())
elif prop_name == 'drop_failed_msg':
    return prop_value.format(actor="You", subject=this.title())
else:
    raise ValueError(f"Unknown property name: {prop_name}")
